/*
 * L.java
 *
 * Created on 25. November 2007, 21:18
 */

/*
 The contents of this file are subject to the
 Mozilla Public License Version 1.1 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied.
 See the License for the specific language governing rights
 and limitations under the License.

 The Initial Developer of the Original Code is
 Georg-Johann Lay (GeorgJohann@web.de).

 Portions created by the Initial Developer are
 Copyright (C) 2007 Georg-Johann Lay. All Rights Reserved.

 Alternatively, the contents of this file may be used under the terms
 of either the GNU General Public License Version 3 or later (the "GPL"),
 or the GNU Lesser General Public License Version 3 or later (the "LGPL"),
 in which case the provisions of the GPL or the LGPL are applicable instead
 of those above.

 If you wish to allow use of your version of this file only
 under the terms of either the GPL or the LGPL, and not to allow others to
 use your version of this file under the terms of the MPL, indicate your
 decision by deleting the provisions above and replace them with the notice
 and other provisions required by the LGPL or the GPL. If you do not delete
 the provisions above, a recipient may use your version of this file under
 the terms of any one of the MPL, the GPL or the LGPL.
 */

package de.gjlay.mobile.localize;

import de.gjlay.mobile.tools.Map;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Hashtable;
import javax.microedition.lcdui.StringItem;

import de.gjlay.mobile.tools.StringTools;

/**
 * This class manages the retrieval of localized strings found in the
 * bundle <code>_MESSAGES_BUNDLE</code>.
 */
public class L
{
    private L() {}
    /**
     * Full path to the messages resource bundle.
     */
    private static final String _MESSAGES_BUNDLE = "/res/strings/messages.properties";

    /**
     * Error message used in the case there is any problem with initialization of
     * localization support.
     */
    private static final String _LOCALIZATION_ERROR_MSG = "Error initializing localization support: ";

    /**
     * Default String is returned from _() methods when there is any problem
     * with finding the appropriate localized message or any part of it.
     */
    private static final String _DEFAULT_STRING = "???";

    /**
     * Initializes localization support based on currently set locale (obtained
     * from system property <code>"microedition.locale"</code>). The initialization method is called
     * automatically when a call to {@link #_(java.lang.String)} method is attempted for the first time.
     *
     * You can call this method explicitly to see whether there is any problem
     * with initialization of the localization support.
     *
     * @return true if the intialization was succesfull, false if there was any problem.
     *         In case of a problem you can get an error message by calling {@link #getErrorMessage()}.
     */
    public static boolean setDefaultLocale ()
    {
        return setLocale (System.getProperty ("microedition.locale"));     // NOI18N
    }

    /**
     * Explicit initialization of the localization support. This method is usually
     * called when a particular locale is wanted.
     * Calling <CODE>setDefaultLocale("fr")</CODE>, e.g., makes the class read messages
     * from <CODE>/res/messages_fr.properties</CODE>.
     *
     * @param locale locale which will be used to determine which message file from bundle will be used
     * @return true if the intialization was succesfull, false if there was any problem.
     *         In case of a problem you can get an error message by calling {@link #getErrorMessage()}.
     */
    public static boolean setLocale (String locale)
    {
        try
        {
            InputStream in = L.getBundleInputStream (locale, _MESSAGES_BUNDLE);
            // load messages to _messageTable hashtable
            if (null != in)
            {
                _messageTable = loadMessages (in);
                return true;
            }
        }
        catch (Exception e)
        {
            // houston we have a problem
            _localizationErrorMessage = _LOCALIZATION_ERROR_MSG + e.getMessage ();
        }

        return false;
    }


/*    
    public static String loadString (String locale, String bundle)
    {
        InputStream in = _getInputStream (locale, bundle);

        if (null != in)
        {
            try
            {
                // load string
                return _loadString (in);
            }
            catch (Exception e)
            {
                // houston we have a problem
                _localizationErrorMessage = _LOCALIZATION_ERROR_MSG + e.getMessage ();
            }
        }

        return _DEFAULT_STRING+" "+bundle+"["+locale+"]";
    }
*/

    /**
     * Returns an error message if there was any problem with accessing the localized
     * text. The message also possibly explainins a reason of the failure. The message
     * is taken from <CODE>_LOCALIZATION_ERROR_MSG</CODE>.
     *
     * @return error message if there was any failure or null when everything is OK.
     */
    public static String getErrorMessage ()
    {
        return _localizationErrorMessage;
    }

    /**
     * Finds a localized string in a message bundle.
     * @param key key of the localized string to look for
     * @return the localized string. If key is not found, then  <CODE>_DEFAULT_STRING</CODE> string
     * is returned
     */
    public static final String _ (String key)
    {
        if (_messageTable == null
                && !setDefaultLocale ())
                return _DEFAULT_STRING;

        String s = (String) _messageTable.get (key);
        return (s == null) ? key : s;
    }

    /**
     * Characters separating keys and values
     */
    private static final String _KEY_VALUE_SEPARATORS = "=: \t\r\n\f";

    /**
     * Characters strictly separating keys and values
     */

    private static final String _STRICT_KEY_VALUE_SEPARTORS = "=:";
    /**
     * white space characters understood by the support (these can be in the message file)
     */
    private static final String _WHITESPACE_CHARS = " \t\r\n\f";


    /**
     * Contains the parsed message bundle.
     */
    private static Hashtable _messageTable;
    /**
     * Contains an error message if there was any problem with localization support.
     * If everything is OK, this field is <code>null</code>.
     */
    private static String _localizationErrorMessage = null;

    /**
     * Loads a string from input stream.
     * @param inStream stream from which the string is read
     * @throws IOException if there is any problem with reading the string
     */
    private static synchronized String _loadString (InputStream inStream)
    throws IOException
    {
        InputStreamReader in = new InputStreamReader (inStream);
        StringBuffer textBuffer = new StringBuffer ();
        while (true)
        {
            // Get next line
            String line = StringTools.readLine (in);
            if (line == null)
                return textBuffer.toString();

            if (0 == line.length())
                textBuffer.append ('\n');
            else
            {
                if (textBuffer.length() > 0)
                    textBuffer.append (' ');
                textBuffer.append (StringTools.convertString(line));
            }
        } // while
    }

    /**
     * Get the InputStream that corresponds to a file within a message bundle.
     *
     * @param locale locale which will be used to determine which message file from bundle will be used
     * @param bundle message bundle to attach the input stream to
     * @return on success return a non-null input stream, <code>null</code> otherwise.
     */
    public static InputStream getBundleInputStream (String locale, String bundle)
    {
        InputStream in = null;
        // need access to a class object - cannot use Object.class, because of MIDP1 bug
        Class clazz = Runtime.getRuntime ().getClass ();
        
        // try to find localized resource first (in format ${name}_locale.${suffix})
        if ((locale != null) && (locale.length () > 1))
        {
            int lastIndex = bundle.lastIndexOf ('.');
            String prefix = bundle.substring (0,lastIndex);
            String suffix = bundle.substring (lastIndex);
            // replace '-' with '_', some phones returns locales with
            // '-' instead of '_'. For example Nokia or Motorola
            locale = locale.replace ('-','_');
            in = clazz.getResourceAsStream (prefix+"_"+locale+suffix);
            if (in == null)
            {
                // if no localized resource is found or localization is available
                // try more general locale (i.e. "en" instead of "en_US").
                in = clazz.getResourceAsStream (prefix+"_"+locale.substring (0,2)+suffix);
            }
        }
        if (in == null)
        {
            // if not found or locale is not set, try default locale
            in = clazz.getResourceAsStream (bundle);
        }

        return in;
    }
    
    /**
     * Loads messages from bundle for given locale
     * @param locale the locale
     * @param bundle the message bundle
     * @return a hashtable that maps keys to values
     * @throws IOException if there is any problem with reading the messages
     */
    public static Hashtable loadMessages (String locale, String bundle)
    throws IOException
    {
        return loadMessages (getBundleInputStream (locale, bundle));
    }
    /**
     * Loads messages from input stream to hash table.
     * @param inStream stream from which the messages are read
     * @return a hashtable that maps keys to values
     * @throws IOException if there is any problem with reading the messages
     */
    public static Hashtable loadMessages (InputStream inStream)
    throws IOException
    {
        final Hashtable messageTable = new Hashtable ();
        
        loadMessages (inStream, new Map()
        {
            public void put (String key, String value)
            {
                messageTable.put (key, value);
            }
        });
        
        return messageTable;
    }
    /**
     * Loads messages from input stream to a map.
     * @param inStream stream from which the messages are read
     * @param map the map that maps keys to values
     * @throws IOException if there is any problem with reading the messages
     */
    public static synchronized void loadMessages (InputStream inStream, Map map)
    throws IOException
    {
        InputStreamReader in = new InputStreamReader (inStream);
        while (true)
        {
            // Get next line
            String line = StringTools.readLine (in);
            if (line == null)
                return;

            if (line.length () > 0)
            {
                // Find start of key
                int len = line.length ();
                int keyStart;
                for (keyStart = 0; keyStart < len; keyStart++)
                    if (_WHITESPACE_CHARS.indexOf (line.charAt (keyStart)) == -1)
                        break;

                // Blank lines are ignored
                if (keyStart == len)
                    continue;

                // Continue lines that end in slashes if they are not comments
                char firstChar = line.charAt (keyStart);
                if ((firstChar != '#') && (firstChar != '!'))
                {
                    while (StringTools.continueLine (line))
                    {
                        String nextLine = StringTools.readLine (in);
                        if (nextLine == null)
                            nextLine = "";
                        String loppedLine = line.substring (0, len - 1);
                        // Advance beyond whitespace on new line
                        int startIndex;
                        for (startIndex = 0; startIndex < nextLine.length (); startIndex++)
                            if (_WHITESPACE_CHARS.indexOf (nextLine.charAt (startIndex)) == -1)
                                break;
                        nextLine = nextLine.substring (startIndex, nextLine.length ());
                        line = new String (loppedLine + nextLine);
                        len = line.length ();
                    }

                    // Find separation between key and value
                    int separatorIndex;
                    for (separatorIndex = keyStart; separatorIndex < len; separatorIndex++)
                    {
                        char currentChar = line.charAt (separatorIndex);
                        if (currentChar == '\\')
                            separatorIndex++;
                        else if (_KEY_VALUE_SEPARATORS.indexOf (currentChar) != -1)
                            break;
                    }

                    // Skip over whitespace after key if any
                    int valueIndex;
                    for (valueIndex = separatorIndex; valueIndex < len; valueIndex++)
                        if (_WHITESPACE_CHARS.indexOf (line.charAt (valueIndex)) == -1)
                            break;

                    // Skip over one non whitespace key value separators if any
                    if (valueIndex < len)
                        if (_STRICT_KEY_VALUE_SEPARTORS.indexOf (line.charAt (valueIndex)) != -1)
                            valueIndex++;

                    // Skip over white space after other separators if any
                    while (valueIndex < len)
                    {
                        if (_WHITESPACE_CHARS.indexOf (line.charAt (valueIndex)) == -1)
                            break;
                        valueIndex++;
                    }
                    String key = line.substring (keyStart, separatorIndex);
                    String value = (separatorIndex < len) ? line.substring (valueIndex, len) : "";

                    // Convert then store key and value
                    key = StringTools.convertString (key);
                    value = StringTools.convertString (value);
                    map.put (key, value);
                }
            }
        }
    }
}
